This is the [Coq](http://coq.inria.fr) development associated to the
[ICFP'20](https://icfp20.sigplan.org/) paper entitled
*"Stable Relations and Abstract Interpretation of Higher-Order Programs"*,
by [Benoît Montagu](https://people.irisa.fr/Benoit.Montagu/) and
[Thomas Jensen](https://www.irisa.fr/celtique/jensen/).

This development is distributed under the terms of the
[CeCILL-B](https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
license, a copy of the license is in the [`LICENSE`](./LICENSE) file.
The files from the `Metalib` directory are distributed under the
[MIT](https://opensource.org/licenses/mit-license.php) license.

# Build instructions

The development is known to compile with [Coq](http://coq.inria.fr) 8.11.0.

- To compile the development, run:

    ```
    make
    ```

    Or, if you want to use 4 parallel jobs, for example, run:

    ```
    make -j 4
    ```

- To run `coqchk` and print the assumptions of the development, run:

    ```
    make validate
    ```

  Notice that *abstract types* created by sealing Coq modules appear
  in the output of `coqchk` as assumptions, although they are
  realised. They should not be considered as axioms.

- To build the documentation in HTML form, run:

    ```
    make doc
    ```

  You can then load the `html/toc.html` file.

- To remove compilation objects, run:

    ```
    make clean
    ```

# Organisation of the development

  The main definitions and lemmas comprise a comment that refers to
  the corresponding definitions, figures, or lemmas of the ICFP'20
  paper. To get a comprehensive list, please search for the string
  "ICFP'20" in the development, using the `grep` tool for instance.

  The development is organised as follows:

  * Documentation files and directory
    - [`README.md`](./README.md): this `README` file
    - [`html/`](./html/): directory that contains the HTML formatted
      documentation. It is obtained by running `make doc`.

  * Main definitions and infrastructure for terms
    - [`Definitions.v`](./Definitions.v): Main definitions (terms,
      operational semantics)
    - [`Metalib/`](./Metalib/):
      [Metalib library](https://github.com/plclub/metalib)
    - [`Infrastructure.v`](./Infrastructure.v): Infrastructure results,
      related to the locally nameless encoding of terms
    - [`StepsFacts.v`](./StepsFacts.v): Results about the small-step semantics
    - [`EvalFacts.v`](./EvalFacts.v): Results about the evaluation relation
    - [`Env.v`](./Env.v): Definitions and proofs about environments
    - [`SubstEnv.v`](./SubstEnv.v): Definitions and proofs about
      substitutions

  * Relations: their definitions and properties
    - [`Rel.v`](./Rel.v): Basic relations
    - [`RelExtra.v`](./RelExtra.v): More relation definitions
    - [`RelLc.v`](./RelLc.v): Results about locally-closed relations
      (i.e. that contain only locally-closed terms, in the sense of the
      locally nameless representation)
    - [`RelClosed.v`](./RelClosed.v): Results about closed relations
      (i.e. that contain only closed terms)
    - [`RelValue.v`](./RelValue.v): Results about value relations (i.e.
      that contain only values)
    - [`RelStable.v`](./RelStable.v): Results about stable relations
    - [`RelGood.v`](./RelGood.v): Results about good relations
    - [`RelWf.v`](./RelWf.v): Wellformed relations

  * Management of names
    - [`TranspCore.v`](./TranspCore.v): Core definitions and results
      about transpositions
    - [`Transp.v`](./Transp.v): More definitions and results about
      transpositions
    - [`EquivarianceFacts.v`](./EquivarianceFacts.v): Equivariance
      properties
    - [`SupportFacts.v`](./SupportFacts.v): Properties about support
    - [`Perm.v`](./Perm.v): Permutations of names, i.e. sequences of
      name transpositions (useful for proving strong induction
      principles that permit to choose names for bound variables)

  * Relation-based semantics of terms
    - [`Rsem.v`](./Rsem.v): Definitions and proofs about the semantics
      of terms as stable relations, based on the small-step operational
      semantics
    - [`OkEnv.v`](./OkEnv.v): Interpretation of environments as sets of
      substitutions
    - [`Interp.v`](./Interp.v): Interpretation of terms as stable
      relations (definition, soundness, completeness)
    - [`InterpStrong.v`](./InterpStrong.v): Strong version of the
      definition of the interpretation (provides stronger induction
      principles, allowing to choose names for bound variables)

  * Relation-based program logics
    - [`RLogic.v`](./RLogic.v): Program logic for stable relations
      (definition, soundness, completeness)
    - [`RLogicStrong.v`](./RLogicStrong.v): Strong version of the
      definition of the program logic (provides stronger induction
      principles, allowing to choose names for bound variables)

  * Examples
    - [`InterpExamples.v`](./InterpExamples.v): Some examples of
      interpretations of λ-terms.

  * Pointwise relations (i.e. relations that pointwise relate
    substitutions with values)
    - [`PointwiseRel.v`](./PoinwisedRel.v): Definitions and proofs for
      poinwise relations

  * Typing of terms (used by correlations only)
    - [`Types.v`](./Types.v): Simple types, typing judgement,
      preservation, progress

  * Correlations (i.e. abstract binary relations on values)
    - [`Corr.v`](./Corr.v): Definitions and proofs about the correlation
      abstract domain
